"""
ToolMate AI Plugin - ask Ollama Chat

Ask Ollama Chat for information

[TOOL_CALL]
"""

from toolmate import config
#from toolmate.utils.ollama_models import ollama_models
from toolmate.utils.call_ollama import CallOllama

def ask_ollama(function_args):
    config.stopSpinning()
    query = function_args.get("query") # required
    config.currentMessages[-1] = {"role": "user", "content": query}
    completion = CallOllama.regularCall(config.currentMessages, chat_model=config.ollamaChatModel if config.useAdditionalChatModel else None)
    config.toolmate.streamCompletion(completion, openai=False)
    if config.useAdditionalChatModel and not config.ollamaChatModel == config.ollamaToolModel:
        CallOllama.unloadModels(config.ollamaChatModel)
    return ""

functionSignature = {
    "examples": [
        "Ask Ollama",
    ],
    "name": "ask_ollama",
    "description": "Ask an Ollama model to chat or provide information",
    "parameters": {
        "type": "object",
        "properties": {
            #"model": {
            #    "type": "string",
            #    "description": "The LLM model name, e.g. 'llama3.1', 'llama3.1:8b', etc.",
            #},
            "query": {
                "type": "string",
                "description": "The request in detail, including any supplementary information",
            },
        },
        "required": ["query"],
    },
}

config.addFunctionCall(signature=functionSignature, method=ask_ollama)
config.inputSuggestions.append("Ask Ollama: ")


"""
ToolMate AI Plugin - ask chatgpt

Ask ChatGPT for conversation only; no function calling

[TOOL_CALL]
"""

if not config.isTermux:

    from toolmate import config, print1, loadLlamacppChatModel
    from toolmate.utils.call_llamacpp import CallLlamaCpp

    def ask_llamacpp(function_args):
        chatModel = None
        config.stopSpinning()
        query = function_args.get("query") # required
        config.currentMessages[-1] = {"role": "user", "content": query}
        if config.useAdditionalChatModel:
            chatModel = loadLlamacppChatModel()
            completion = CallLlamaCpp.regularCall(config.currentMessages, model=chatModel)
        else:
            completion = CallLlamaCpp.regularCall(config.currentMessages)
        config.toolmate.streamCompletion(completion, openai=False)
        if chatModel is not None:
            try:
                chatModel.close()
                print1("Llama.cpp chat model unloaded!")
            except:
                pass
        return ""

    functionSignature = {
        "examples": [
            "Ask Llama.cpp",
        ],
        "name": "ask_llamacpp",
        "description": "Ask Llama.cpp to chat or provide information",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=ask_llamacpp)
    config.inputSuggestions.append("Ask ChatGPT: ")


"""
ToolMate AI Plugin - build agents

build a group of agents to execute a task with integrated "AutoGen Agent Builder"

[TOOL_CALL]
"""


if not config.isTermux:
    try:
        from toolmate.autobuilder import AutoGenBuilder
        from toolmate import config
        from toolmate import print1, print2, print3

        def build_agents(function_args):
            if not config.openaiApiKey or config.openaiApiKey == "toolmate":
                return "OpenAI API key not found! This feature works with ChatGPT models only!"
            task = function_args.get("task") # required
            title = function_args.get("title") # required
            print2("AutoGen Agent Builder launched!")
            print3(f"Title: {title}")
            print3(f"Description: {task}")
            messages = AutoGenBuilder().getResponse(task, title)
            if not messages[-1]["content"]:
                del messages[-1]
            # add context to the message chain
            config.currentMessages += messages
            print2("\nAutoGen Agent Builder closed!")
            return ""

        functionSignature = {
            "examples": [
                "autobuilder",
                "create a team of assistants",
                "create a crew of agents",
            ],
            "name": "build_agents",
            "description": "build a group of AI assistants or agents to execute a complicated task that other functions cannot resolve",
            "parameters": {
                "type": "object",
                "properties": {
                    "task": {
                        "type": "string",
                        "description": "Task description in as much detail as possible",
                    },
                    "title": {
                        "type": "string",
                        "description": "A short title to describe the task",
                    },
                },
                "required": ["task", "title"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=build_agents)
    except:
        print("Plugin `create ai assistants` not enabled! Run `pip install pyautogen[autobuild]` first!")
        pass


"""
ToolMate AI Plugin - edit text

edit text files

[TOOL_CALL]
"""

from toolmate import config

def chat(_):
    # a dummy plugin to avoid using tool
    return "[INVALID]"

functionSignature = {
    "examples": [
        "Hi",
        "How are you?",
        "In python, how"
    ],
    "name": "chat",
    "description": f'''Provide information or answer a question''',
    "parameters": {
        "type": "object",
        "properties": {
            "message": {
                "type": "string",
                "description": "My message",
            },
        },
        "required": ["message"],
    },
}

config.addFunctionCall(signature=functionSignature, method=chat)


"""
ToolMate AI Plugin - improve English

add context to help users to improve English
"""

from toolmate import config

config.predefinedChatSystemMessages["British English Teacher"] = """You're here to be my friendly British English teacher. Your goal is to assist me in enhancing my language skills, whether it's improving my grammar, refining my speaking style, or selecting the appropriate words. You're dedicated to supporting my progress and together, we can collaborate on enhancing my British English. Let's engage in a conversation, in the style of British spoken English, and imagine that we're based in London. Feel free to initiate discussions about anything related to London or the UK. Remember, you want to chat with me like a real friend, so instead of just giving me lots of information, let's have a random talk and you'll help me improve my English as we chat. Please always strive to maintain an engaging and continuous conversation."""

config.predefinedChatSystemMessages["Spoken English Teacher"] = """I want you to act as a spoken English teacher and improver. I will speak to you in English and you will reply to me in English to practice my spoken English. I want you to keep your reply neat, limiting the reply to 100 words. I want you to strictly correct my grammar mistakes, typos, and factual errors. I want you to ask me a question in your reply. Now let's start practicing, you could ask me a question first. Remember, I want you to strictly correct my grammar mistakes, typos, and factual errors."""

config.predefinedChatSystemMessages["English Translator and Improver"] = """I want you to act as an English translator, spelling corrector and improver. I will speak to you in any language and you will detect the language, translate it and answer in the corrected and improved version of my text, in English. I want you to replace my simplified A0-level words and sentences with more beautiful and elegant, upper level English words and sentences. Keep the meaning same, but make them more literary. I want you to only reply the correction, the improvements and nothing else, do not write explanations."""


"""
ToolMate AI Plugin - ask Tavily

Ask Tavily

[TOOL_CALL]
"""

if not config.isTermux:

    from toolmate import config, getTavilyClient, print1

    def ask_tavily(function_args):
        query = function_args.get("query") # required
        config.currentMessages[-1] = {"role": "user", "content": query}
        config.toolTextOutput = getTavilyClient().qna_search(query=query)
        config.stopSpinning()
        print1(config.toolTextOutput)
        return ""

    functionSignature = {
        "examples": [
            "Ask internet",
        ],
        "name": "ask_tavily",
        "description": "Ask internet for a short and direct answer",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=ask_tavily)
    config.inputSuggestions.append("Ask Tavily: ")


"""
ToolMate AI Plugin - memory

Save and retrieve memory

modified from source: https://medium.com/@simon_attard/building-a-memory-layer-for-gpt-using-function-calling-da17d66920d0

[TOOL_CALL]
"""

from toolmate import config, get_or_create_collection, add_vector, query_vectors
from toolmate import print1
from pathlib import Path
from chromadb.config import Settings
import os, chromadb, getpass, geocoder, datetime, json

memory_store = os.path.join(config.localStorage, "memory")
Path(memory_store).mkdir(parents=True, exist_ok=True)
chroma_client = chromadb.PersistentClient(memory_store, Settings(anonymized_telemetry=False))

#import numpy as np
#from numpy.linalg import norm
#def cosine_similarity(A, B):
#    cosine = np.dot(A, B) / (norm(A) * norm(B))
#    return cosine


if not config.isTermux:

    def save_memory(function_args):
        memory = function_args.get("memory") # required
        memory_title = function_args.get("title") # required
        memory_type = function_args.get("type") # required
        memory_tags = function_args.get("tags") # required
        if not isinstance(memory_tags, str):
            memory_tags = str(memory_tags)
        collection = get_or_create_collection(chroma_client, "memories")
        g = geocoder.ip('me')
        metadata = {
            "timestamp": str(datetime.datetime.now()),
            "tags": memory_tags,
            "title": memory_title,
            "type": memory_type,
            "user": getpass.getuser(),
            "location": f"{g.city}, {g.state}, {g.country}",
        }
        if config.developer:
            print1(config.divider)
            print(">>> saving memory: ")
            print1(f"memory: {memory}")
            print(metadata)
            print1(config.divider)
        add_vector(collection, memory, metadata)
        config.stopSpinning()
        return "I saved it in my memory!"

    def search_memory(function_args):
        query = function_args.get("query") # required
        collection = get_or_create_collection(chroma_client, "memories")
        res = query_vectors(collection, query, config.memoryClosestMatches)
        if config.developer:
            print1(config.divider)
            print(">>> retrieved memories: ") 
            print(res["documents"])
            print1(config.divider)
        info = {}
        for index, description in enumerate(res["documents"][0]):
            info[f"memory {index}"] = {
                "description": description,
            }
        config.stopSpinning()
        return json.dumps(info)

    functionSignature1 = {
        "examples": [
            "remember, ",
            "save in your memory",
            "keep in mind",
            "memorize",
            "make a note",
        ],
        "name": "save_memory",
        "description": """Use this function if I mention something which you think would be useful in the future and should be saved as a memory. Saved memories will allow you to retrieve snippets of past conversations when needed.""",
        "parameters": {
            "type": "object",
            "properties": {
                "memory": {
                    "type": "string",
                    "description": "Full description of the memory to be saved. I would like you to help me with converting relative dates and times, if any, into exact dates and times based on the given current date and time.",
                },
                "title": {
                    "type": "string",
                    "description": "Title of the memory",
                },
                "type": {
                    "type": "string",
                    "description": "Type of the memory, return either 'general', 'instruction', 'fact', 'event', or 'concept'",
                    "enum": ["general", "instruction", "fact", "event", "concept"],
                },
                "tags": {
                    "type": "string",
                    "description": """Return a list of tags about the memory, e.g. '["work", "to_do", "follow_up"]'"""
                },
            },
            "required": ["memory", "title", "type", "tags"]
        }
    }
    functionSignature2 = {
        "examples": [
            "search your memory",
            "recall from your memory",
            "do you remember",
        ],
        "name": "search_memory",
        "description": """Recall memories of important conversation snippets that we had in the past.""",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The query to be used to look up memories from a vector database"
                },
            },
            "required": ["query"]
        }
    }

    config.inputSuggestions += ["Remember, ", "Do you remember?"]
    config.addFunctionCall(signature=functionSignature1, method=save_memory, datetimeSensitive=True)
    config.addFunctionCall(signature=functionSignature2, method=search_memory, datetimeSensitive=True)


"""
ToolMate AI Plugin - add a calendar event

add a calendar event on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config, stopSpinning

    def add_calendar_event(function_args):
        title = function_args.get("title") # required
        description = function_args.get("description") # required
        url = function_args.get("url", "") # optional
        start_time = function_args.get("start_time_converted_in_milliseconds", "") # optional
        end_time = function_args.get("end_time_converted_in_milliseconds", "") # optional
        location = function_args.get("location", "") # optional

        insert_url = f"\nURL: {url}\n" if url else ""
        insert_location = f"\nLocation: {location}" if location else ""
        description = f'''{description}{insert_url}{insert_location}'''.replace('"', '\\"')

        stopSpinning()

        cli = f'''am start -a android.intent.action.INSERT -t vnd.android.cursor.item/event -e title "{title}" -e description "{description}" -e beginTime {start_time} -e endTime {end_time} -e location {location}'''
        subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        return ""

    functionSignature = {
        "examples": [
            "add a calendar event",
        ],
        "name": "add_calendar_event",
        "description": "Add a calendar event",
        "parameters": {
            "type": "object",
            "properties": {
                "title": {
                    "type": "string",
                    "description": "The title of the event.",
                },
                "description": {
                    "type": "string",
                    "description": "The detailed description of the event, including the people involved and their roles, if any.",
                },
                "url": {
                    "type": "string",
                    "description": "Event url",
                },
                "start_time_converted_in_milliseconds": {
                    "type": "string",
                    "description": "The start date and time converted in milliseconds since epoch",
                },
                "end_time_converted_in_milliseconds": {
                    "type": "string",
                    "description": "The start date and time converted in milliseconds since epoch. If not given, return 1 hour later than the start_time_converted_in_milliseconds",
                },
                "location": {
                    "type": "string",
                    "description": "The location or venue of the event.",
                },
            },
            "required": ["title", "description"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=add_calendar_event, deviceInfo=True, datetimeSensitive=True)


"""
This plugin works with optional module `bible`, install it by:

> pip install toolmate[bible]
"""

from toolmate import config
from toolmate import print1, print2, print3, print4, removeDuplicatedListItems, stopSpinning, openURL
from toolmate.utils.text_utils import TextUtil
from toolmate.utils.regex_search import RegexSearch
from flashtext import KeywordProcessor
import traceback, re, requests, os

persistentConfigs = (
    ("uniquebible_api_endpoint", "https://bible.gospelchurch.uk/plain"),
    ("uniquebible_api_timeout", 10),
    ("uniquebible_api_private", ""),
    ("uniquebible_weburl", "https://bible.gospelchurch.uk/index.html"),
)
config.setConfig(persistentConfigs)
temporaryConfigs = (
    ("uniquebible", False),
    ("uniquebible_platform", None),
)
config.setConfig(temporaryConfigs, temporary=True)

# Tool: @uniquebible_api @bapi
try:

    private = f"private={config.uniquebible_api_private}&" if config.uniquebible_api_private else ""
    r = requests.get(f"{config.uniquebible_api_endpoint}?{private}cmd=.suggestions", timeout=config.uniquebible_api_timeout)
    r.encoding = "utf-8"
    apiCommandSuggestions = r.json()

    def uniquebible_api(_):
        stopSpinning()

        private = f"private={config.uniquebible_api_private}&" if config.uniquebible_api_private else ""
        command = config.currentMessages[-1]["content"] #.replace('"', '\\"')

        url = f"""{config.uniquebible_api_endpoint}?{private}cmd={command}"""
        response = requests.get(url, timeout=config.uniquebible_api_timeout)
        response.encoding = "utf-8"
        config.toolTextOutput = response.text.strip()
        
        print2("\n```UniqueBible API")
        print1(config.toolTextOutput)
        print2("```")

        return ""
    functionSignature = {
        "examples": [
            "@uniquebible_api BIBLE:::NET:::John 3:16",
            "@uniquebible_api CROSSREFERENCE:::John 3:16",
        ],
        "name": "uniquebible_api",
        "description": "Retrieve bible data with UniqueBible API",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    }
    config.addFunctionCall(signature=functionSignature, method=uniquebible_api)
    config.aliases["@bapi "] = "@uniquebible_api "
    config.builtinTools["bapi"] = "Retrieve bible data with UniqueBible API"
    config.inputSuggestions.append({"@uniquebible_api": apiCommandSuggestions})
    config.inputSuggestions.append({"@bapi": apiCommandSuggestions})

    # Tool: @uniquebible_web
    def uniquebible_web(_):
        stopSpinning()
        command = config.currentMessages[-1]["content"].replace('"', '\\"')
        url = f"""{config.uniquebible_weburl}?cmd={command}"""
        openURL(url)
        return ""
    functionSignature = {
        "examples": [
            "@uniquebible_web BIBLE:::NET:::John 3:16",
            "@uniquebible_web CROSSREFERENCE:::John 3:16",
        ],
        "name": "uniquebible_web",
        "description": "Read bible-related content via UniqueBible web interface",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    }
    config.addFunctionCall(signature=functionSignature, method=uniquebible_web)
    config.inputSuggestions.append({"@uniquebible_web": apiCommandSuggestions})

except:
    print(f"Failed to connect '{config.uniquebible_api_endpoint}' at the moment!")

try:
    # load resources information
    cwd = os.getcwd()
    # change to UBA user content directory temporarily
    ubaUserDir = os.path.join(os.path.expanduser("~"), "UniqueBible")
    os.chdir(ubaUserDir)

    from uniquebible.util.ConfigUtil import ConfigUtil
    ConfigUtil.setup(noQt=True, runMode="terminal")
    from uniquebible import config as bibleconfig
    #print(bibleconfig.packageDir, bibleconfig.ubaUserDir)
    from uniquebible.util.BibleBooks import BibleBooks
    from uniquebible.util.CrossPlatform import CrossPlatform
    from uniquebible.util.BibleVerseParser import BibleVerseParser
    from uniquebible.util.LocalCliHandler import LocalCliHandler
    from uniquebible.db.BiblesSqlite import Bible
    from uniquebible.db.ToolsSqlite import Commentary

    config.uniquebible_platform = CrossPlatform()
    config.uniquebible_platform.setupResourceLists()
    config.uniquebible_localCliHandler = LocalCliHandler()
    os.chdir(cwd)

    """ available resources:
    config.uniquebible_platform.textList
    config.uniquebible_platform.textFullNameList
    config.uniquebible_platform.strongBibles
    config.uniquebible_platform.commentaryList
    config.uniquebible_platform.commentaryFullNameList
    config.uniquebible_platform.referenceBookList
    config.uniquebible_platform.topicListAbb
    config.uniquebible_platform.topicList
    config.uniquebible_platform.lexiconList
    config.uniquebible_platform.dictionaryListAbb
    config.uniquebible_platform.dictionaryList
    config.uniquebible_platform.encyclopediaListAbb
    config.uniquebible_platform.encyclopediaList
    config.uniquebible_platform.thirdPartyDictionaryList
    config.uniquebible_platform.pdfList
    config.uniquebible_platform.epubList
    config.uniquebible_platform.docxList
    config.uniquebible_platform.bibleAudioModules
    config.uniquebible_platform.dataList
    config.uniquebible_platform.searchToolList
    """

    # Tool: @extract_bible_references
    def extract_bible_references(_):
        stopSpinning()
        content = config.currentMessages[-1]["content"]
        config.toolTextOutput = BibleVerseParser(False).extractAllReferencesReadable(content)
        print2("\n```references")
        print1(config.toolTextOutput if config.toolTextOutput else "[not found]")
        print2("```")
        return ""
    functionSignature = {
        "examples": [
            "@extract_bible_references I like John 3:16",
        ],
        "name": "extract_bible_references",
        "description": "Extract Bible references from a block of text",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    }
    config.addFunctionCall(signature=functionSignature, method=extract_bible_references)
    config.inputSuggestions.append("Extract Bible references: ")

    # Tool: @bible
    def bible(_):
        stopSpinning()
        # change to uniquebible app directory temporarily
        cwd = os.getcwd()
        os.chdir(bibleconfig.ubaUserDir)

        def displayBibleVerses(verseList, text=None, searchString=""):
            bible = Bible(text=text)
            if searchString and not verseList:
                # perform a regex / plain text search if no bible reference is found
                # use tools @search_bible or @search_bible_paragraphs for more advanced searches
                try:
                    # regex search
                    query = "SELECT Book, Chapter, Verse FROM Verses WHERE (Scripture REGEXP ?)"
                    binding = (searchString,)
                    fullVerseList = bible.getSearchVerses(query, binding)
                except:
                    # plain text search, in case an invalid regex pattern is given
                    query = "SELECT Book, Chapter, Verse FROM Verses WHERE (Scripture LIKE ?)"
                    binding = (f"%{searchString}%",)
                    fullVerseList = bible.getSearchVerses(query, binding)
            else:
                # include all verses within a range
                fullVerseList = bible.getEverySingleVerseList(verseList)
            for verse in fullVerseList:
                b, c, v, verseText = bible.readTextVerse(*verse, noAudioTag=True)
                verseText = re.sub("<[^<>]*?>", "", verseText)
                verseText = f"({parser.bcvToVerseReference(b, c, v)}) {verseText}"
                print4(verseText)
                config.toolTextOutput += f"{verseText}\n"

        # display verses
        config.toolTextOutput = ""
        content = config.currentMessages[-1]["content"]
        content = re.sub("([0-9]) (:[0-9])", r"\1\2", content)
        parser = BibleVerseParser(True)
        verseList = parser.extractAllReferences(content)

        keyword_processor = KeywordProcessor()
        keyword_processor.add_keywords_from_list([f"`{i}`" for i in config.uniquebible_platform.textList])
        bibles = removeDuplicatedListItems(keyword_processor.extract_keywords(content))
        if not bibles:
            bibles = [f"`{bibleconfig.mainText}`"]
        if not verseList:
            content = re.sub("|".join(bibles), "", content)
            content = re.sub(" [ ]+?([^ ])", r" \1", content).rstrip()
        for i in bibles:
            i = i[1:-1]
            heading = f"# Bible: {i}"
            config.toolTextOutput += f"\n{heading}\n\n"
            print()
            print2(heading)
            displayBibleVerses(verseList, text=i, searchString="" if verseList else content)
        # save the last opened bible as the default bible
        bibleconfig.mainText = bibles[-1][1:-1]
        ConfigUtil.save()

        config.toolTextOutput = config.toolTextOutput.strip()

        # change back to previous directory
        os.chdir(cwd)
        return ""
    functionSignature = {
        "examples": [
            "@bible John 3:16-18", # provide a single reference; use default bible version if bible version is not given
            "@bible `NIV` John 3:16-18; Deu 6:4", # specify a bible version and multiple references
            "@bible `NIV` `ESV` John 3:16-18; Deu 6:4", # specify multiple bible versions and multiple references
            "@bible Jesus love", # perform a plain text search
            "@bible `NET` apostle of Christ", # search for 'apostle of Christ' in NET bible
        ],
        "name": "bible",
        "description": "Retrieve Bible verses based on given references or perform a plain text search",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    }
    config.addFunctionCall(signature=functionSignature, method=bible)
    config.inputSuggestions.append("Show bible verses: ")

    # input suggestions
    standardBibleBooks = []
    bookSuggestions = {}
    abbrev = BibleBooks.abbrev["eng"]
    for i in range(1, 67):
        abb, fullName = abbrev[str(i)]
        standardBibleBooks.append(abb)
        standardBibleBooks.append(fullName)
        chapters = BibleBooks.chapters[i]
        bookSuggestions[abb] = {str(ii): {f":{iii}": None for iii in range(1, BibleBooks.verses[i][ii]+1)} for ii in range(1, chapters+1)}
        bookSuggestions[fullName] = {str(ii): {f":{iii}": None for iii in range(1, BibleBooks.verses[i][ii]+1)} for ii in range(1, chapters+1)}
    config.inputSuggestions.append(bookSuggestions)
    bibleSuggestions = {}
    for i in config.uniquebible_platform.textList:
        bibleSuggestions[f"`{i}`"] = bookSuggestions
    for i in standardBibleBooks:
        bibleSuggestions[i] = bookSuggestions[i]
    config.inputSuggestions.append({"@bible": bibleSuggestions})

    # Tool: @uniquebible @uba
    def uniquebible(_):
        stopSpinning()
        # change to uniquebible app directory temporarily
        cwd = os.getcwd()
        os.chdir(bibleconfig.ubaUserDir)

        command = config.currentMessages[-1]["content"].replace('"', '\\"')

        rawOutput = bibleconfig.rawOutput
        bibleconfig.rawOutput = True
        addFavouriteToMultiRef = bibleconfig.addFavouriteToMultiRef
        bibleconfig.addFavouriteToMultiRef = False # disable addFavouriteToMultiRef for the output
        config.toolTextOutput = config.uniquebible_localCliHandler.getContent(command, False).strip()
        bibleconfig.addFavouriteToMultiRef = addFavouriteToMultiRef
        bibleconfig.rawOutput = rawOutput
        # alternative: loading stream mode; slower
        #config.toolTextOutput = subprocess.run(f'''uniquebible stream "{command}"''', shell=True, capture_output=True, text=True).stdout.strip()
        print2("\n```UniqueBible App")
        print1(config.toolTextOutput)
        print2("```")
        
        os.chdir(cwd)
        return ""
    functionSignature = {
        "examples": [
            "@uniquebible BIBLE:::NET:::John 3:16",
            "@uniquebible CROSSREFERENCE:::John 3:16",
        ],
        "name": "uniquebible",
        "description": "Retrieve bible data with UniqueBible App commands",
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    }
    config.addFunctionCall(signature=functionSignature, method=uniquebible)
    config.aliases["@b "] = "@uniquebible "
    config.builtinTools["b"] = "Retrieve bible data with UniqueBible App commands"

    textCommandSuggestion = [key + ":::" for key in config.uniquebible_localCliHandler.textCommandParser.interpreters.keys()]
    commandCompleterSuggestions = config.uniquebible_localCliHandler.getCommandCompleterSuggestions(textCommandSuggestion=textCommandSuggestion)
    config.inputSuggestions.append({"@uniquebible": commandCompleterSuggestions})
    config.inputSuggestions.append({"@b": commandCompleterSuggestions})

    # Tool: @bible_commentary
    if config.uniquebible_platform.commentaryList:
        def bible_commentary(_):
            stopSpinning()
            # change to uniquebible app directory temporarily
            cwd = os.getcwd()
            os.chdir(bibleconfig.ubaUserDir)

            def getCommentaryContent(module, verse):
                content = ""
                commentary = Commentary(module)
                b, c, v, *_ = verse
                query = "SELECT Scripture FROM Commentary WHERE Book=? AND Chapter=?"
                commentary.cursor.execute(query, verse[0:2])
                chapterCommentary = commentary.cursor.fetchone()[0]
                if chapterCommentary:
                    pattern = '(<vid id="v[0-9]+?.[0-9]+?.[0-9]+?"></vid>)<hr>'
                    searchReplaceItems = ((pattern, r"<hr>\1"),)
                    chapterCommentary = RegexSearch.deepReplace(chapterCommentary, pattern, searchReplaceItems)
                    verseCommentaries = chapterCommentary.split("<hr>")
                    for i in verseCommentaries:
                        if f'<vid id="v{b}.{c}.{v}"' in i:
                            content = i
                            break
                return content
            # display verses
            config.toolTextOutput = ""
            content = config.currentMessages[-1]["content"]
            content = re.sub("([0-9]) (:[0-9])", r"\1\2", content)
            parser = BibleVerseParser(True)
            verseList = parser.extractAllReferences(content)
            bible = Bible(text="KJV")
            fullVerseList = bible.getEverySingleVerseList(verseList)

            keyword_processor = KeywordProcessor()
            keyword_processor.add_keywords_from_list([f"`{i}`" for i in config.uniquebible_platform.commentaryList])
            commentaries = removeDuplicatedListItems(keyword_processor.extract_keywords(content))
            if not commentaries:
                commentaries = [f"`{bibleconfig.commentaryText}`"]

            for i in commentaries:
                loaded = []
                i = i[1:-1]
                for ii in fullVerseList:
                    content = getCommentaryContent(i, ii)
                    if content and not content in loaded:
                        loaded.append(content)
                        heading = f"# Commentary: {i} - {parser.bcvToVerseReference(*ii)}"
                        config.toolTextOutput += f"\n{heading}\n\n"
                        print()
                        print2(heading)
                        content = TextUtil.htmlToPlainText(content)
                        config.toolTextOutput += f"{content}\n"
                        try:
                            print1(content)
                        except:
                            print(content)
            # save the last opened commentary as the default commentary
            bibleconfig.commentaryText = commentaries[-1][1:-1]
            ConfigUtil.save()

            config.toolTextOutput = config.toolTextOutput.strip()

            # change back to previous directory
            os.chdir(cwd)
            return ""
        functionSignature = {
            "examples": [
                "@bible_commentary John 3:16",
                "@bible_commentary `CBSC` `BKC` John 3:16-18; Deu 6:4",
            ],
            "name": "bible_commentary",
            "description": "Retrieve bible commentary",
            "parameters": {
                "type": "object",
                "properties": {},
                "required": [],
            },
        }
        config.addFunctionCall(signature=functionSignature, method=bible_commentary)
        config.inputSuggestions.append("Read bible commentary: ")

        commentarySuggestions = {}
        for i in config.uniquebible_platform.commentaryList:
            commentarySuggestions[f"`{i}`"] = bookSuggestions
        for i in standardBibleBooks:
            commentarySuggestions[i] = bookSuggestions[i]
        config.inputSuggestions.append({"@bible_commentary": commentarySuggestions})

    # searchbible.ai
    try:
        cwd = os.getcwd()
        from searchbible.searchbible import search
        from searchbible import config as searchbibleconfig
        config.searchbible_path = searchbibleconfig.packageFolder
        os.chdir(cwd)
    except:
        config.searchbible_path = ""

    if config.searchbible_path:
        # Tool: @search_bible
        def search_bible(_):
            stopSpinning()
            content = config.currentMessages[-1]["content"]
            print2("\n```search_bible")
            config.toolTextOutput = "```search_bible\n" + search(simpleSearch="" if content == "[NONE]" else content) + "\n```"
            print2("```")
            return ""
        functionSignature = {
            "examples": [
                "@search_bible David fled", # perform a simple similarity search for 'David fled'
                "@search_bible", # when search item is not specified, users are prompted to customise available search parameters.
            ],
            "name": "search_bible",
            "description": "Perform similarity search for verses in the bible",
            "parameters": {
                "type": "object",
                "properties": {},
                "required": [],
            },
        }
        config.addFunctionCall(signature=functionSignature, method=search_bible)

        # Tool: @search_bible_paragraphs
        def search_bible_paragraphs(_):
            stopSpinning()
            content = config.currentMessages[-1]["content"]
            print2("\n```search_bible_paragraphs")
            config.toolTextOutput = "```search_bible_paragraphs\n" + search(simpleSearch="" if content == "[NONE]" else content) + "\n```"
            print2("```")
            return ""
        functionSignature = {
            "examples": [
                "@search_bible_paragraphs David fled", # perform a simple similarity search for 'David fled'
                "@search_bible_paragraphs", # when search item is not specified, users are prompted to customise available search parameters.
            ],
            "name": "search_bible_paragraphs",
            "description": "Perform similarity search for paragraphs in the bible",
            "parameters": {
                "type": "object",
                "properties": {},
                "required": [],
            },
        }
        config.addFunctionCall(signature=functionSignature, method=search_bible_paragraphs)
    config.uniquebible = True
except:
    #print(traceback.format_exc())
    config.uniquebible = False

"""
Notes about other bible-related API:

CUV Bible
Reference: https://bible.fhl.net/json
cuv = requests.get("https://bible.fhl.net/json/qb.php?gb=0&chap=3&sec=6&chineses=出")
cuv.text.encode().decode('unicode_escape')

NET Bible
Reference: https://labs.bible.org/
https://labs.bible.org/api/?passage=John+3:16-17;%20Deut%206:4
"""


import os
from toolmate import config, print2, print3, getUnstructuredFiles, get_or_create_collection, add_vector, query_vectors, refinePath, ragSearchContext, ragRefineDocsPath, ragGetSplits, getRagPrompt
from toolmate.utils.call_llm import CallLLM
import os, traceback, re, datetime, traceback, threading, shutil, chromadb
from chromadb.config import Settings
from pathlib import Path
from toolmate.utils.prompts import Prompts
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.styles import Style
from autogen.retrieve_utils import TEXT_FORMATS
# from langchain_community.embeddings.sentence_transformer import SentenceTransformerEmbeddings
# from langchain_community.embeddings import HuggingFaceEmbeddings
from toolmate.utils.streaming_word_wrapper import StreamingWordWrapper


class RAG:

    def __init__(self):
        if not hasattr(config, "currentMessages"):
            CallLLM.checkCompletion()

        # alternative
        """
        rag_store = os.path.join(config.localStorage, "rag")
        try:
            shutil.rmtree(rag_store, ignore_errors=True)
            #print2("Old retrieval store removed!")
        except:
            print2("Failed to remove old rag store!")
        Path(rag_store).mkdir(parents=True, exist_ok=True)
        chroma_client = chromadb.PersistentClient(rag_store, Settings(anonymized_telemetry=False))
        self.collection = get_or_create_collection(chroma_client, "rag")

    def getContext2(self, splits, query):
        # add to vector database
        for i in splits:
            add_vector(self.collection, i.page_content, metadata={"source": i.metadata.get("source")[0]})
        
        # search
        retrieved_docs = query_vectors(self.collection, query, n=config.rag_closestMatches)["documents"]
        if not retrieved_docs:
            return None

        # format context
        #formatted_context = "\n\n".join([i.page_content for i in retrieved_docs])
        formatted_context = {f"information_{index}": content for index, content in enumerate(retrieved_docs)}
        #formatted_context = [i[0] for i in retrieved_docs]
        return formatted_context"""

    def getResponse(self, docs_path, query):
        docs_path = ragRefineDocsPath(docs_path)
        if docs_path is None:
            return None
        splits = ragGetSplits(docs_path)
        formatted_context = ragSearchContext(splits, query)
        if not formatted_context:
            with open(docs_path[0], "r", encoding="utf-8") as fileObj:
                formatted_context = fileObj.read()

        # format prompt
        formatted_prompt = getRagPrompt(query, formatted_context)

        if hasattr(config, "currentMessages"):
            messages = config.currentMessages[:-1] + [{"role": "user", "content" : formatted_prompt}]
        else:
            messages = [{"role": "user", "content" : formatted_prompt}]

        #completion = CallLLM.regularCall(messages)
        #config.toolmate.streamCompletion(completion)

        try:
            streamingWordWrapper = StreamingWordWrapper()
            completion = CallLLM.regularCall(messages)
            # Create a new thread for the streaming task
            streaming_event = threading.Event()
            self.streaming_thread = threading.Thread(target=streamingWordWrapper.streamOutputs, args=(streaming_event, completion, True if config.llmInterface in ("chatgpt", "letmedoit", "groq", "llamacppserver") else False))
            # Start the streaming thread
            self.streaming_thread.start()
            # wait while text output is steaming; capture key combo 'ctrl+q' or 'ctrl+z' to stop the streaming
            streamingWordWrapper.keyToStopStreaming(streaming_event)
            # when streaming is done or when user press "ctrl+q"
            self.streaming_thread.join()
        except:
            print(traceback.format_exc())
            self.streaming_thread.join()

        return ""

    def print(self, message):
        #print(message)
        print_formatted_text(HTML(message))

    def run(self):
        promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        prompts = Prompts()

        print2(f"Retrieval utility launched!")
        self.print(f"""[press '{str(config.hotkey_exit).replace("'", "")[1:-1]}' to exit]""")
        
        print2(f"Enter document path below (file / folder):")
        self.print(f"""Supported formats: *.{", *.".join(TEXT_FORMATS)}""" + ", *.zip")
        docs_path = prompts.simplePrompt(style=promptStyle)

        # handle path dragged to terminal
        docs_path = refinePath(docs_path)

        if docs_path and os.path.exists(docs_path):
            self.print("Enter your query below:")
            query = prompts.simplePrompt(style=promptStyle)
            if not query == config.exit_entry:
                try:
                    self.getResponse(docs_path, query)
                except:
                    print(traceback.format_exc())
        else:
            print2(f"Entered path does not exist!")
        
        print2(f"Retrieval utility closed!")

def main():
    RAG().run()

if __name__ == '__main__':
    main()


from toolmate import config, getCliOutput
import pydoc, os, re, sys
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.shortcuts import set_title

prompt_shared_key_bindings = KeyBindings()

# selection

# select / unselect all
@prompt_shared_key_bindings.add("c-a")
def _(event):
    buffer = event.app.current_buffer
    if config.selectAll:
        text = buffer.text
        buffer.reset()
        buffer.text = text
        buffer.cursor_position = len(buffer.text)
    else:
        buffer.cursor_position = 0
        buffer.start_selection()
        buffer.cursor_position = len(buffer.text)
    config.selectAll = not config.selectAll

# clipboard

# copy text to clipboard
@prompt_shared_key_bindings.add("c-c")
def _(event):
    buffer = event.app.current_buffer
    data = buffer.copy_selection()
    copyText = data.text
    if config.terminalEnableTermuxAPI:
        pydoc.pipepager(copyText, cmd="termux-clipboard-set")
    else:
        # remarks: set_data does not work
        config.clipboard.set_text(copyText)
# paste clipboard text
@prompt_shared_key_bindings.add("c-v")
def _(event):
    buffer = event.app.current_buffer
    buffer.cut_selection()
    if config.terminalEnableTermuxAPI:
        clipboardText = getCliOutput("termux-clipboard-get")
    else:
        clipboardText = config.clipboard.get_data().text
    buffer.insert_text(clipboardText)
# cut text to clipboard
@prompt_shared_key_bindings.add("c-x")
def _(event):
    buffer = event.app.current_buffer
    data = buffer.cut_selection()
    # remarks: set_data does not work
    config.clipboard.set_text(data.text)
# insert linebreak
@prompt_shared_key_bindings.add(*config.hotkey_insert_newline)
def _(event):
    buffer = event.app.current_buffer
    if config.multilineInput:
        buffer.validate_and_handle()
    else:
        buffer.newline()

# navigation

# go to current line starting position
@prompt_shared_key_bindings.add("escape", "b")
def _(event):
    buffer = event.app.current_buffer
    buffer.cursor_position = buffer.cursor_position - buffer.document.cursor_position_col
# go to current line ending position
@prompt_shared_key_bindings.add("escape", "e")
def _(event):
    buffer = event.app.current_buffer
    buffer.cursor_position = buffer.cursor_position + buffer.document.get_end_of_line_position()
# go to current line starting position
@prompt_shared_key_bindings.add("home")
def _(event):
    buffer = event.app.current_buffer
    buffer.cursor_position = buffer.cursor_position - buffer.document.cursor_position_col
# go to current line ending position
@prompt_shared_key_bindings.add("end")
def _(event):
    buffer = event.app.current_buffer
    buffer.cursor_position = buffer.cursor_position + buffer.document.get_end_of_line_position()
# go to the end of the text
@prompt_shared_key_bindings.add("escape", "z")
def _(event):
    buffer = event.app.current_buffer
    buffer.cursor_position = len(buffer.text)
# go to the beginning of the text
@prompt_shared_key_bindings.add("escape", "a")
def _(event):
    buffer = event.app.current_buffer
    buffer.cursor_position = 0

# reset buffer
@prompt_shared_key_bindings.add(*config.hotkey_cancel)
def _(event):
    buffer = event.app.current_buffer
    buffer.reset()

# open current input in built-in or custom text editor
# place this method here so that terminal or other input can share this binding
#@prompt_shared_key_bindings.add("escape", "o")
@prompt_shared_key_bindings.add(*config.hotkey_edit_current_entry)
def _(event):
    customTextEditor = config.customTextEditor if config.customTextEditor else f"{sys.executable} {os.path.join(config.toolMateAIFolder, 'eTextEdit.py')}"
    current_buffer = event.app.current_buffer
    text = current_buffer.text
    filename = os.path.join(config.toolMateAIFolder, "temp", "current_input.txt")
    with open(filename, "w", encoding="utf-8") as fileObj:
        fileObj.write(text)
    os.system(f"{customTextEditor} {filename}")
    with open(filename, "r", encoding="utf-8") as fileObj:
        editedText = fileObj.read()
    editedText = re.sub("\n$", "", editedText)
    current_buffer.text = editedText
    current_buffer.cursor_position = len(editedText)
    set_title(config.toolMateAIName)

# swap color theme
@prompt_shared_key_bindings.add(*config.hotkey_swap_text_brightness)
def _(_):
    swapTerminalColors()

@staticmethod
def swapTerminalColors():
    if config.terminalResourceLinkColor in config.terminalColors:
        config.terminalResourceLinkColor = config.terminalColors[config.terminalResourceLinkColor]
    if config.terminalPromptIndicatorColor1 in config.terminalColors:
        config.terminalPromptIndicatorColor1 = config.terminalColors[config.terminalPromptIndicatorColor1]
    if config.terminalCommandEntryColor1 in config.terminalColors:
        config.terminalCommandEntryColor1 = config.terminalColors[config.terminalCommandEntryColor1]
    if config.terminalPromptIndicatorColor2 in config.terminalColors:
        config.terminalPromptIndicatorColor2 = config.terminalColors[config.terminalPromptIndicatorColor2]
    if config.terminalCommandEntryColor2 in config.terminalColors:
        config.terminalCommandEntryColor2 = config.terminalColors[config.terminalCommandEntryColor2]
    if config.terminalHeadingTextColor in config.terminalColors:
        config.terminalHeadingTextColor = config.terminalColors[config.terminalHeadingTextColor]
#    if config.terminalVerseNumberColor in config.terminalColors:
#        config.terminalVerseNumberColor = config.terminalColors[config.terminalVerseNumberColor]
#    if config.terminalVerseSelectionBackground in config.terminalColors:
#        config.terminalVerseSelectionBackground = config.terminalColors[config.terminalVerseSelectionBackground]
#    if config.terminalVerseSelectionForeground in config.terminalColors:
#        config.terminalVerseSelectionForeground = config.terminalColors[config.terminalVerseSelectionForeground]
#    if config.terminalSearchHighlightBackground in config.terminalColors:
#        config.terminalSearchHighlightBackground = config.terminalColors[config.terminalSearchHighlightBackground]
#    if config.terminalSearchHighlightForeground in config.terminalColors:
#        config.terminalSearchHighlightForeground = config.terminalColors[config.terminalSearchHighlightForeground]
#    if config.terminalFindHighlightBackground in config.terminalColors:
#        config.terminalFindHighlightBackground = config.terminalColors[config.terminalFindHighlightBackground]
#    if config.terminalFindHighlightForeground in config.terminalColors:
#        config.terminalFindHighlightForeground = config.terminalColors[config.terminalFindHighlightForeground]
    #config.terminalSwapColors = (config.terminalResourceLinkColor.startswith("ansibright"))

# edit
# insert spaces by pressing the SHIFT+TAB key
@prompt_shared_key_bindings.add("s-tab")
def _(event):
    buffer = event.app.current_buffer
    buffer.insert_text(config.terminalEditorTabText)
# backspace
@prompt_shared_key_bindings.add("c-h")
def _(event):
    buffer = event.app.current_buffer
    data = buffer.cut_selection()
    # delete one char before cursor [backspace] if there is no text selection
    if not data.text and buffer.cursor_position >= 1:
        buffer.delete_before_cursor(1)
# forward delete
@prompt_shared_key_bindings.add("c-d")
def _(event):
    buffer = event.app.current_buffer
    data = buffer.cut_selection()
    # forward delete one character if there is no selection
    if not data.text and buffer.cursor_position < len(buffer.text):
        buffer.delete(1)
# replace selection
@prompt_shared_key_bindings.add("<any>")
def _(event):
    buffer = event.app.current_buffer
    buffer.cut_selection().text
    # a key sequence looks like [KeyPress(key='a', data='a')]
    buffer.insert_text(event.key_sequence[0].data)



"""
ToolMate AI Plugin - create images

generate images with model "dall-e-3"

[TOOL_CALL]
"""

if not config.isTermux:

    from toolmate import config, print2, print3, getCurrentDateTime, getCliOutput, getCpuThreads, downloadStableDiffusionFiles
    import os, shutil
    from base64 import b64decode
    from toolmate.utils.call_chatgpt import check_openai_errors
    from toolmate.utils.terminal_mode_dialogs import TerminalModeDialogs
    from openai import OpenAI
    from pathlib import Path
    from stable_diffusion_cpp import StableDiffusion
    from toolmate.utils.single_prompt import SinglePrompt
    from prompt_toolkit.styles import Style
    from toolmate.utils.prompt_validator import NumberValidator


    def create_image_sd(function_args):
        def callback(step: int, steps: int, time: float):
            print("Completed step: {} of {}".format(step, steps))

        def openImageFile(imageFile):
            if config.terminalEnableTermuxAPI:
                getCliOutput(f"termux-share {imageFile}")
            elif shutil.which(config.open):
                cli = f"{config.open} {imageFile}"
                #os.system(cli)
                subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            message = f"Image saved: {imageFile}"
            config.toolTextOutput = message
            print3(message)

        prompt = function_args.get("prompt") # required

        # image file path
        folder = os.path.join(config.localStorage, "images")
        Path(folder).mkdir(parents=True, exist_ok=True)
        imageFile = os.path.join(folder, f"{getCurrentDateTime()}.png")

        config.stopSpinning()

        # customize width and height
        promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        change = False
        print("# Width & Height")
        print2("Specify the width:")
        new_width = SinglePrompt.run(style=promptStyle, default=str(config.stableDiffusion_output_width), validator=NumberValidator())
        if new_width and not new_width.strip().lower() == config.exit_entry and int(new_width) > 0 and not new_width == config.stableDiffusion_output_width:
            config.stableDiffusion_output_width = int(new_width)
            change = True
        print2("Specify the height:")
        new_height = SinglePrompt.run(style=promptStyle, default=str(config.stableDiffusion_output_height), validator=NumberValidator())
        if new_height and not new_height.strip().lower() == config.exit_entry and int(new_height) > 0 and not new_height == config.stableDiffusion_output_height:
            config.stableDiffusion_output_height = int(new_height)
            change = True
        print("# Sample steps")
        print1("Increasing the number of sampling steps generally enhances image quality by refining details and reducing noise, but it also requires more processing time.")
        print2("Specify the sample steps:")
        new_stableDiffusion_sample_steps = SinglePrompt.run(style=promptStyle, default=str(config.stableDiffusion_sample_steps), validator=NumberValidator())
        if new_stableDiffusion_sample_steps and not new_stableDiffusion_sample_steps.strip().lower() == config.exit_entry and int(new_stableDiffusion_sample_steps) > 0 and not new_stableDiffusion_sample_steps == config.stableDiffusion_sample_steps:
            config.stableDiffusion_sample_steps = int(new_stableDiffusion_sample_steps)
            change = True
        if change:
            config.saveConfig()

        downloadStableDiffusionFiles()
        stable_diffusion = StableDiffusion(
            model_path=config.stableDiffusion_model_path,
            lora_model_dir=os.path.join(config.localStorage, "LLMs", "stable_diffusion", "lora"),
            wtype="default", # Weight type (options: default, f32, f16, q4_0, q4_1, q5_0, q5_1, q8_0)
            # seed=1337, # Uncomment to set a specific seed
            verbose=config.stableDiffusion_verbose,
            n_threads=getCpuThreads(),
        )
        stable_diffusion.txt_to_img(
            prompt,
            width=config.stableDiffusion_output_width,
            height=config.stableDiffusion_output_height,
            sample_steps=config.stableDiffusion_sample_steps,
            progress_callback=callback,
        )[0].save(imageFile)
        openImageFile(imageFile)
        return ""


    functionSignature = {
        "examples": [
            "generate image",
            "create image",
        ],
        "name": "create_image_sd",
        "description": "Create an image with Stable Diffusion Models",
        "parameters": {
            "type": "object",
            "properties": {
                "prompt": {
                    "type": "string",
                    "description": "Description of the image in as much detail as possible",
                },
            },
            "required": ["prompt"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=create_image_sd)


from toolmate import config, showErrors, isValidPythodCode, executeToolFunction, toParameterSchema, getCpuThreads, useChatSystemMessage
from toolmate import print1, print2, print3, getPythonFunctionResponse, extractPythonCode, encode_image
from typing import Optional
from llama_cpp import Llama
from llama_cpp.llama_chat_format import Llava15ChatHandler
from prompt_toolkit import prompt
from pathlib import Path
from huggingface_hub import hf_hub_download
import traceback, json, re, os, pprint, copy, datetime


class CallLlamaCpp:

    @staticmethod
    def checkCompletion():

        try:
            config.llamacppToolModel.close()
            print1("Llama.cpp model unloaded!")
        except:
            pass

        llm_directory = os.path.join(config.localStorage, "LLMs", "gguf")
        Path(llm_directory).mkdir(parents=True, exist_ok=True)

        def loadMainModel():
            llamacppToolModel_model_path = os.path.join(llm_directory, config.llamacppToolModel_filename)
            if not config.llamacppToolModel_model_path or not os.path.isfile(config.llamacppToolModel_model_path):
                config.llamacppToolModel_model_path = llamacppToolModel_model_path
            if not os.path.isfile(config.llamacppToolModel_model_path):
                # download llava clip model
                print2("Downloading tool model ...")
                hf_hub_download(
                    repo_id=config.llamacppToolModel_repo_id,
                    filename=config.llamacppToolModel_filename,
                    local_dir=llm_directory,
                    #local_dir_use_symlinks=False,
                )
            config.llamacppToolModel = None
            cpuThreads = getCpuThreads()
            config.llamacppToolModel = Llama(
                model_path=config.llamacppToolModel_model_path,
                chat_format="chatml",
                n_ctx=config.llamacppToolModel_n_ctx,
                n_batch=config.llamacppToolModel_n_batch,
                verbose=config.llamacppToolModel_verbose,
                n_threads=cpuThreads,
                n_threads_batch=cpuThreads,
                n_gpu_layers=config.llamacppToolModel_n_gpu_layers,
                **config.llamacppToolModel_additional_model_options,
            )

        def downloadChatModel():
            llamacppChatModel_model_path = os.path.join(llm_directory, config.llamacppChatModel_filename)
            if not config.llamacppChatModel_model_path or not os.path.isfile(config.llamacppChatModel_model_path):
                config.llamacppChatModel_model_path = llamacppChatModel_model_path
            if not os.path.isfile(config.llamacppChatModel_model_path):
                # download llava clip model
                print2("Downloading chat model ...")
                hf_hub_download(
                    repo_id=config.llamacppChatModel_repo_id,
                    filename=config.llamacppChatModel_filename,
                    local_dir=llm_directory,
                    #local_dir_use_symlinks=False,
                )

        # Download main and code models
        try:
            loadMainModel()
        except:
            # restore default config
            print2("Errors! Restoring default tool model!")
            config.llamacppToolModel_ollama_tag = ""
            config.llamacppToolModel_model_path = ""
            config.llamacppToolModel_repo_id = "MaziyarPanahi/WizardLM-2-7B-GGUF"
            config.llamacppToolModel_filename = "WizardLM-2-7B.Q4_K_M.gguf"
            loadMainModel()

        try:
            downloadChatModel()
        except:
            # restore default config
            print2("Errors! Restoring default chat model!")
            config.llamacppChatModel_ollama_tag = ""
            config.llamacppChatModel_model_path = ""
            config.llamacppChatModel_repo_id = "bartowski/Meta-Llama-3.1-8B-Instruct-GGUF"
            config.llamacppChatModel_filename = "Meta-Llama-3.1-8B-Instruct-Q4_K_M.gguf"
            downloadChatModel()

        # Download vision model
        filename = "ggml-model-q4_k.gguf"
        llamacppVisionModel_model_path = os.path.join(llm_directory, filename)
        if not config.llamacppVisionModel_model_path or not os.path.isfile(config.llamacppVisionModel_model_path):
            config.llamacppVisionModel_model_path = llamacppVisionModel_model_path
        if not os.path.isfile(config.llamacppVisionModel_model_path):
            # download llava model
            print2(f"Downloading llava model ...")
            hf_hub_download(
                repo_id="mys/ggml_llava-v1.5-7b",
                filename=filename,
                local_dir=llm_directory,
                #local_dir_use_symlinks=False,
            )

        filename = "mmproj-model-f16.gguf"
        llamacppVisionModel_clip_model_path = os.path.join(llm_directory, filename)
        if not config.llamacppVisionModel_clip_model_path or not os.path.isfile(config.llamacppVisionModel_clip_model_path):
            config.llamacppVisionModel_clip_model_path = llamacppVisionModel_clip_model_path
        if not os.path.isfile(config.llamacppVisionModel_clip_model_path):
            # download llava clip model
            print2("Downloading clip model ...")
            hf_hub_download(
                repo_id="mys/ggml_llava-v1.5-7b",
                filename=filename,
                local_dir=llm_directory,
                #local_dir_use_symlinks=False,
            )
        
        config.saveConfig()

    @staticmethod
    def autoCorrectPythonCode(code, trace):
        for i in range(config.max_consecutive_auto_correction):
            userInput = f"""I ran this copy of python code:
```
{code}
```

However, I got the following error:
```
{trace}
```

Please generate another copy of python code that fix the error.

Remember, output the new copy of python code ONLY, without additional notes or explanation."""
            messages = [{"role": "user", "content" : userInput}]
            print3(f"Auto-correction attempt: {(i + 1)}")
            function_call_message, function_call_response = CallLlamaCpp.getSingleFunctionCallResponse(messages, "correct_python_code")
            arguments = function_call_message["function_call"]["arguments"]
            if not arguments:
                print2("Generating code ...")
                response = CallLlamaCpp.getSingleChatResponse(userInput, prefill="```python\n", stop=["```"]).replace(r"\\n", "\n")
                python_code = extractPythonCode(response)
                if isValidPythodCode(python_code):
                    arguments = {
                        "code": python_code,
                        "missing": [],
                        "issue": "",
                    }
                    function_call_response = executeToolFunction(arguments, "correct_python_code")
                else:
                    continue

            # display response
            print1(config.divider)
            if config.developer:
                print(function_call_response)
            else:
                print1("Executed!" if function_call_response == "EXECUTED" else "Failed!")
            if function_call_response == "EXECUTED":
                break
            else:
                code = arguments.get("code")
                trace = function_call_response
            print1(config.divider)
        
        # return information if any
        if function_call_response == "EXECUTED":
            pythonFunctionResponse = getPythonFunctionResponse(code)
            if pythonFunctionResponse:
                return json.dumps({"information": pythonFunctionResponse})
            else:
                return ""
        # ask if user want to manually edit the code
        print1(f"Failed to execute the code {(config.max_consecutive_auto_correction + 1)} times in a row!")
        print1("Do you want to manually edit it? [y]es / [N]o")
        confirmation = prompt(style=config.promptStyle2, default="N")
        if confirmation.lower() in ("y", "yes"):
            config.defaultEntry = f"```python\n{code}\n```"
            return ""
        else:
            return "[INVALID]"

    @staticmethod
    def regularCall(messages: dict, temperature: Optional[float]=None, max_tokens: Optional[int]=None, model=None):
        chatMessages = useChatSystemMessage(copy.deepcopy(messages))
        if model is not None:
            return model.create_chat_completion(
                messages=chatMessages,
                temperature=temperature if temperature is not None else config.llmTemperature,
                max_tokens=max_tokens if max_tokens is not None else config.llamacppToolModel_max_tokens,
                stream=True,
                **config.llamacppToolModel_additional_chat_options,
            )            
        return config.llamacppToolModel.create_chat_completion(
            messages=chatMessages,
            temperature=temperature if temperature is not None else config.llmTemperature,
            max_tokens=max_tokens if max_tokens is not None else config.llamacppToolModel_max_tokens,
            stream=True,
            **config.llamacppToolModel_additional_chat_options,
        )

    @staticmethod
    def img_to_txt(file_path: str, query: str="Describe this image in detail please.", temperature: Optional[float]=None, max_tokens: Optional[int]=None, messages: Optional[list]=None):
        chat_handler = Llava15ChatHandler(clip_model_path=config.llamacppVisionModel_clip_model_path)
        cpuThreads = getCpuThreads()
        llm = Llama(
            model_path=config.llamacppVisionModel_model_path,
            chat_handler=chat_handler,
            logits_all=True, # needed to make llava work
            n_ctx=config.llamacppVisionModel_n_ctx, # n_ctx should be increased to accomodate the image embedding
            n_batch=config.llamacppVisionModel_n_batch,
            verbose=config.llamacppVisionModel_verbose,
            n_threads=cpuThreads,
            n_threads_batch=cpuThreads,
            n_gpu_layers=config.llamacppVisionModel_n_gpu_layers,
            **config.llamacppVisionModel_additional_model_options,
        )
        # update messages
        if messages is None:
            messages = [
                {
                    "role": "system",
                    "content": "You are an assistant who perfectly describes images.",
                },
            ]
        messages.append(
            {
                "role": "user",
                "content": [
                    {"type": "image_url", "image_url": {"url": encode_image(file_path)}},
                    {"type" : "text", "text": query},
                ]
            }
        )
        try:
            completion = llm.create_chat_completion(
                messages=messages,
                temperature=temperature if temperature is not None else config.llmTemperature,
                max_tokens=max_tokens if max_tokens is not None else config.llamacppVisionModel_max_tokens,
                stream=False,
                **config.llamacppVisionModel_additional_chat_options,
            )
            return completion["choices"][0]["message"].get("content", "")
        except:
            return ""

    @staticmethod
    def getDictionaryOutput(messages: list, schema: dict={}, temperature: Optional[float]=None, max_tokens: Optional[int]=None) -> dict:
        schema = toParameterSchema(schema)
        try:
            completion = config.llamacppToolModel.create_chat_completion(
                messages=messages,
                response_format={"type": "json_object", "schema": schema} if schema else {"type": "json_object"},
                temperature=temperature if temperature is not None else config.llmTemperature,
                max_tokens=max_tokens if max_tokens is not None else config.llamacppToolModel_max_tokens,
                stream=False,
                **config.llamacppToolModel_additional_chat_options,
            )
            jsonOutput = completion["choices"][0]["message"].get("content", "{}")
            jsonOutput = re.sub("^[^{]*?({.*?})[^}]*?$", r"\1", jsonOutput)
            responseDict = json.loads(jsonOutput)
            return responseDict
        except:
            showErrors()
            return {}

    @staticmethod
    def getSingleChatResponse(userInput: str, messages: list=[], temperature: Optional[float]=None, max_tokens: Optional[int]=None, prefill: Optional[str]=None, stop: Optional[list]=None):
        # non-streaming single call
        if userInput:
            item = {"role": "user", "content" : userInput}
            if messages and messages[-1].get("role", "") == "assistant":
                messages.insert(-1, item)
            else:
                messages.append(item)
        chatMessages = useChatSystemMessage(copy.deepcopy(messages))
        if prefill is not None:
            chatMessages.append({'role': 'assistant', 'content': prefill})
        try:
            completion = config.llamacppToolModel.create_chat_completion(
                messages=chatMessages,
                temperature=temperature if temperature is not None else config.llmTemperature,
                max_tokens=max_tokens if max_tokens is not None else config.llamacppToolModel_max_tokens,
                stop=stop if stop else None,
                stream=False,
                **config.llamacppToolModel_additional_chat_options,
            )
            return completion["choices"][0]["message"].get("content", "")
        except:
            return ""

    # Specific Function call equivalence

    @staticmethod
    def runSingleFunctionCall(messages, function_name):
        messagesCopy = copy.deepcopy(messages)
        try:
            _, function_call_response = CallLlamaCpp.getSingleFunctionCallResponse(messages, function_name)
            function_call_response = function_call_response if function_call_response else config.toolTextOutput
            messages[-1]["content"] += f"""\n\nAvailable information:\n{function_call_response}"""
            config.toolTextOutput = ""
        except:
            showErrors()
            return messagesCopy
        return messages

    @staticmethod
    def getSingleFunctionCallResponse(messages: list, function_name: str, temperature: Optional[float]=None, max_tokens: Optional[int]=None):
        tool_schema = config.toolFunctionSchemas[function_name]["parameters"]
        user_request = messages[-1]["content"]
        func_arguments = CallLlamaCpp.extractToolParameters(schema=tool_schema, userInput=user_request, ongoingMessages=messages, temperature=temperature, max_tokens=max_tokens)
        function_call_response = executeToolFunction(func_arguments=func_arguments, function_name=function_name)
        function_call_message_mini = {
            "role": "assistant",
            "content": "",
            "function_call": {
                "name": function_name,
                "arguments": func_arguments,
            }
        }
        return function_call_message_mini, function_call_response

    # Auto Function call equivalence

    @staticmethod
    def runToolCall(messages: dict):
        user_request = messages[-1]["content"]
        if not config.selectedTool:
            return CallLlamaCpp.regularCall(messages)
        else:
            # 2. Tool Selection
            if config.selectedTool and not config.selectedTool == "chat" and config.selectedTool in config.toolFunctionSchemas:
                tool_name = config.selectedTool
                tool_schema = config.toolFunctionSchemas[tool_name]["parameters"]
                config.selectedTool = ""
            else:
                return CallLlamaCpp.regularCall(messages)
            # 3. Parameter Extraction
            if config.developer:
                print1("extracting parameters ...")
            try:
                if not tool_schema["properties"]:
                    # Execute function directly
                    tool_response = executeToolFunction(func_arguments={}, function_name=tool_name)
                else:
                    tool_parameters = CallLlamaCpp.extractToolParameters(schema=tool_schema, userInput=user_request, ongoingMessages=messages)
                    if not tool_parameters:
                        if config.developer:
                            print1("Failed to extract parameters!")
                        return CallLlamaCpp.regularCall(messages)
                    # 4. Function Execution
                    tool_response = executeToolFunction(func_arguments=tool_parameters, function_name=tool_name)
            except:
                print(traceback.format_exc())
                tool_response = "[INVALID]"
            # 5. Chat Extension
            if tool_response == "[INVALID]":
                # invalid tool call; return a regular call instead
                return CallLlamaCpp.regularCall(messages)
            else:
                # record tool selection
                #config.currentMessages[-1]["tool"] = tool_name
                if tool_response:
                    if config.developer:
                        print2(config.divider)
                        print2("Tool output:")
                        print(tool_response)
                        print2(config.divider)
                    messages[-1]["content"] = f"""Response to the following query according to given supplementary information.

Query:
<query>
{user_request}
</query>

Supplementary information:
<supplementary_information>
{tool_response}
</supplementary_information>"""
                    return CallLlamaCpp.regularCall(messages)
                elif (not config.currentMessages[-1].get("role", "") == "assistant" and not config.currentMessages[-2].get("role", "") == "assistant") or (config.currentMessages[-1].get("role", "") == "system" and not config.currentMessages[-2].get("role", "") == "assistant"):
                    # tool function executed without chat extension
                    config.currentMessages.append({"role": "assistant", "content": config.toolTextOutput if config.toolTextOutput else "Done!"})
                    config.toolTextOutput = ""
                    config.conversationStarted = True
                    return None

    @staticmethod
    def extractToolParameters(schema: dict, userInput: str, ongoingMessages: list = [], temperature: Optional[float]=None, max_tokens: Optional[int]=None) -> dict:
        """
        Extract action parameters
        """
        schema = toParameterSchema(schema)
        schemaCopy = copy.deepcopy(schema)

        # Generate Code when required
        if "code" in schema["required"]:
            del schemaCopy["properties"]["code"]
            schemaCopy["required"].remove("code")
            enforceCodeOutput = """ Remember, you should format the requested information, if any, into a string that is easily readable by humans. Use the 'print' function in the final line to display the requested information."""
            code_instruction = schema["properties"]["code"]["description"] + enforceCodeOutput
            code_instruction = f"""Generate python code according to the following instruction:
</instruction>
{code_instruction}
</instruction>

Here is my request:
<request>
{userInput}
</request>

Remember, response with the required python code ONLY, WITHOUT extra notes or explanations."""

            code = CallLlamaCpp.getSingleChatResponse(code_instruction, ongoingMessages[:-1], temperature, max_tokens, prefill="```python\n", stop=["```"]).replace(r"\\n", "\n")
            code = extractPythonCode(code, keepInvalid=True)
            if len(schema["properties"]) == 1:
                return {"code": code}
        else:
            code = ""

        codeContext = f"""

Find required code below:
{code}""" if code else ""

        messages = ongoingMessages[:-2] + [
            {
                "role": "system",
                "content": f"""You are a JSON builder expert that outputs in JSON.""",
            },
            {
                "role": "user",
                "content": f"""Response in JSON based on the following content:

<content>
{userInput}{codeContext}
</content>

Generate content to fill up the value of each required key in the JSON, if information is not provided.

Remember, output in JSON.""",
            },
        ]

        # schema alternative: properties if len(properties) == 1 else schema
        parameters = CallLlamaCpp.getDictionaryOutput(messages, schemaCopy, temperature=temperature, max_tokens=max_tokens)
        if code:
            parameters["code"] = code

        if config.developer:
            print2("```parameters")
            pprint.pprint(parameters)
            print2("```")
        return parameters


"""
ToolMate AI Plugin - ask chatgpt

Ask ChatGPT for conversation only; no function calling

[TOOL_CALL]
"""


from toolmate import config, getDynamicTokens, print1
#from toolmate.utils.call_chatgpt import CallChatGPT
import copy

try:
    CallChatGPT.checkCompletion()

    def ask_o1_mini(function_args):
        config.stopSpinning()
        query = function_args.get("query") # required
        config.currentMessages[-1] = {"role": "user", "content": query}

        # read beta limitations at https://platform.openai.com/docs/guides/reasoning/beta-limitations
        chatMessages = useChatSystemMessage(copy.deepcopy(config.currentMessages))
        chatMessages = [i for i in chatMessages if not i.get("role", "") == "system"] # beta doesn't support system message
        completion = config.oai_client.chat.completions.create(
            model="o1-mini",
            messages=chatMessages,
            max_completion_tokens=65536,
        )
        config.toolTextOutput = completion.choices[0].message.content
        print1(config.toolTextOutput)
        return ""

    functionSignature = {
        "examples": [
            "Ask o1-mini",
        ],
        "name": "ask_o1_mini",
        "description": "Ask reasoning model o1-mini to chat or provide information",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=ask_o1_mini)
    config.inputSuggestions.append("Ask o1-mini: ")

except:
    print("Plugin `ask o1-mini` not enabled! Check if your OpenAI API key is valid!")


"""
ToolMate AI Plugin - character analysis

add contexts for character analysis

"""

from toolmate import config

config.predefinedContexts["Character - Close Reading"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply 'Close Reading' to analyze the character given in my input. (Remember, Close Reading involves analyzing the text carefully to gather information about the character. It focuses on examining the character's actions, dialogue, thoughts, and interactions with other characters.)
Below is my input:
"""

config.predefinedContexts["Character - Character Analysis"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply 'Character Analysis' to analyze the character given in my input. (Remember, Character Analysis involves examining the character's traits, motivations, and development throughout the narrative. It may include analyzing the character's background, relationships, and conflicts.)
Below is my input:
"""

config.predefinedContexts["Character - Archetypal Analysis"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply 'Archetypal Analysis' to analyze the character given in my input. (Remember, Archetypal Analysis involves identifying and analyzing the character's archetype or symbolic role in the narrative. It explores how the character embodies certain universal themes or patterns.)
Below is my input:
"""

config.predefinedContexts["Character - Psychological Analysis"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply 'Psychological Analysis' to analyze the character given in my input. (Remember, Psychological Analysis involves applying psychological theories and concepts to understand the character's behavior and motivations. It may involve exploring the character's personality, desires, fears, and conflicts.)
Below is my input:
"""

config.predefinedContexts["Character - Historical and Cultural Analysis"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply 'Historical and Cultural Analysis' to analyze the character given in my input. (Remember, Historical and Cultural Analysis involves examining the character in the context of the historical and cultural setting of the narrative. It explores how the character's actions and beliefs may be influenced by their social, political, or cultural environment.)
Below is my input:
"""

config.predefinedContexts["Character - Comparative Analysis"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply 'Comparative Analysis' to analyze the character given in my input. (Remember, Comparative Analysis involves comparing the character to other characters in the narrative or to characters from other texts. It may focus on similarities and differences in their traits, roles, or thematic significance.)
Below is my input:
"""

config.predefinedContexts["Character - Narrative Therapy 1"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply the principles of narrative therapy to analyze the character given in my input. Please explore the character's narrative, beliefs, values, and the meanings they attribute to their experiences. Please also give insights into their identity, relationships, struggles, and potential for growth or change.
Below is my input:
"""

config.predefinedContexts["Character - Narrative Therapy 2"] = """@chat Use the following step-by-step instructions to respond to my inputs.
Step 1: Give me a brief summary of the narrative.
Step 2: Apply principles of Narrative Therapy, that are 1) Externalizing Problems, 2) Re-authoring Stories, 3) Deconstructing Dominant Discourses, 4) Encouraging Unique Outcomes, 5) Externalizing and Privileging Alternative Stories, to analyze the given character in my input.
Below is my input:
"""

config.predefinedContexts["Character - Narrative Therapy 3"] = """@chat Apply narrative therapy to analyze this character below:
"""

"""
The key principles of narrative therapy include:

1. Externalizing Problems: Instead of seeing problems as inherent to the individual, narrative therapy views them as separate entities that can be examined and addressed objectively.

2. Re-authoring Stories: Individuals are encouraged to reframe their narratives and create new meanings and understandings of their experiences. This can help shift focus from problems to strengths and resources.

3. Deconstructing Dominant Discourses: The therapy aims to challenge societal and cultural narratives that may contribute to individuals feeling marginalized or oppressed. By questioning these dominant discourses, individuals can gain a sense of agency and empowerment.

4. Encouraging Unique Outcomes: Narrative therapy seeks to identify and amplify unique outcomes or instances in which the individual has exhibited resilience or problem-solving skills. This helps to counterbalance the dominant problem-focused narratives.

5. Externalizing and Privileging Alternative Stories: The therapist collaborates with the individual to explore alternative stories or narratives that challenge the dominant problem-saturated ones. These alternative stories highlight the individual's strengths, values, and preferred ways of being.

These principles guide the practice of narrative therapy and help individuals to explore and transform their stories in ways that promote healing and growth.
"""


"""
ToolMate AI Plugin - search chat records

search and open old chat records

[TOOL_CALL]
"""

if not config.isTermux:

    from toolmate import config, get_or_create_collection, add_vector, query_vectors, showErrors, displayLoadedMessages
    from toolmate import print1, print2, print3
    from pathlib import Path
    from chromadb.config import Settings
    import os, chromadb, re

    chat_store = os.path.join(config.localStorage, "chats")
    Path(chat_store).mkdir(parents=True, exist_ok=True)
    chroma_client = chromadb.PersistentClient(chat_store, Settings(anonymized_telemetry=False))


    def save_chat_record(timestamp, order, record):
        role = record.get("role", "")
        content = record.get("content", "")
        tool = record.get("tool", "")
        if role and role in ("user", "assistant") and content:
            collection = get_or_create_collection(chroma_client, "chats")
            metadata = {
                "platform": config.llmInterface,
                "timestamp": timestamp,
                "order": order,
                "role": role,
                "tool": tool,
            }
            add_vector(collection, content, metadata)
    config.save_chat_record = save_chat_record

    def search_conversations(function_args):
        query = function_args.get("query") # required
        print3(f"""Query: {query}""")
        collection = get_or_create_collection(chroma_client, "chats")
        res = query_vectors(collection, query, config.chatRecordClosestMatches)
        config.stopSpinning()
        if res:
            exampleID = ""
            # display search results
            print2(config.divider)
            print(">>> retrieved chat records: ")
            for metadata, document in zip(res["metadatas"][0], res["documents"][0]):
                print1(config.divider)
                print3(f"""Chat ID: {metadata["timestamp"]}""")
                if not exampleID:
                    exampleID = metadata["timestamp"]
                print3(f"""Order: {metadata["order"]}""")
                print3(f"""Role: {metadata["role"]}""")
                print3(f"""Content: {document}""")
            print1(config.divider)
            print2("Tips: You can load old chat records by quoting a chat ID or timestamp, e.g.")
            print1(f">>> Load chat records with this ID: {exampleID}")
            print2(config.divider)
        return ""

    def load_conversations(function_args):

        def validateChatFile(chatFile):
            chatFile = os.path.expanduser(chatFile)
            if os.path.isfile(chatFile):
                isfile = True
            elif re.search("^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]_[0-9][0-9]_[0-9][0-9]_[0-9][0-9]$", chatFile):
                # match chat id format
                folderPath = os.path.join(config.localStorage, "chats", re.sub("^([0-9]+?-[0-9]+?)-.*?$", r"\1", chatFile))
                chatFile = os.path.join(folderPath, f"{chatFile}.txt")
                if os.path.isfile(chatFile):
                    isfile = True
                else:
                    isfile = False
            else:
                isfile = False
            return (isfile, chatFile)

        config.stopSpinning()
        timestamp = function_args.get("id") # required
        isfile, chatFile = validateChatFile(timestamp)
        if not isfile:
            print3(f"Invalid chat ID / file path: {timestamp}")
            return "[INVALID]"

        print3(f"Loading chat records: {timestamp} ...")

        try:
            with open(chatFile, "r", encoding="utf-8") as fileObj:
                messages = fileObj.read()
            currentMessages = eval(messages)
            if type(currentMessages) == list:
                config.currentMessages = [{"content": i.get("content", ""), "role": "user"} if i.get("role", "") == "user" and config.llmInterface in ("chatgpt", "letmedoit", "groq", "llamacppserver") else i for i in currentMessages] # make sure "tool" is not in user message
                # display loaded messages
                displayLoadedMessages(config.currentMessages)
                return ""
            else:
                print3(f"Failed to load chat records '{timestamp}' due to invalid format!")
        except:
            print3(f"Failed to load chat records: {timestamp}\n")
            showErrors()
        return "[INVALID]"

    functionSignature1 = {
        "examples": [
            "search our conversations",
            "search chat records",
            "find in chat history",
        ],
        "name": "search_conversations",
        "description": """Search chat records or conversations""",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The search query in detail"
                },
            },
            "required": ["query"]
        }
    }

    functionSignature2 = {
        "examples": [
            "load chat record",
            "open chat history",
        ],
        "name": "load_conversations",
        "description": """Load a saved conversations if chat ID / timestamp / file path is given""",
        "parameters": {
            "type": "object",
            "properties": {
                "id": {
                    "type": "string",
                    "description": "The chat ID or timestamp or a file path"
                },
            },
            "required": ["id"]
        }
    }

    config.inputSuggestions += ["Search chat records: ", "Load chat records with this ID: ", "Load chat records in this file: "]
    config.addFunctionCall(signature=functionSignature1, method=search_conversations)
    config.addFunctionCall(signature=functionSignature2, method=load_conversations)


"""
ToolMate AI Plugin - search financial data

search financial data with yfinance

[TOOL_CALL]
"""

if not config.isTermux:

    from toolmate import config
    from toolmate.utils.python_utils import PythonUtil
    import json

    def search_finance(function_args):
        code = function_args.get("code") # required
        information = PythonUtil.showAndExecutePythonCode(code)
        if information:
            info = {
                "information": information,
            }
            return json.dumps(info)
        return ""

    functionSignature = {
        "examples": [
            "check stock price",
        ],
        "name": "search_finance",
        "description": f'''Search or analyze financial data. Use this function ONLY WHEN package yfinance is useful to resolve my request''',
        "parameters": {
            "type": "object",
            "properties": {
                "code": {
                    "type": "string",
                    "description": "Generate python code that integrates package yfinance to resolve my request. Integrate package matplotlib to visualize data, if applicable.",
                },
            },
            "required": ["code"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=search_finance)


"""
ToolMate AI Plugin - make phone call

make a phone call on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config
    import subprocess

    def phone_call(function_args):
        phone_number = function_args.get("phone_number", "") # required
        cli = f'''termux-telephony-call {phone_number}'''
        subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return ""

    functionSignature = {
        "examples": [],
        "name": "phone_call",
        "description": f'''Make a phone call''',
        "parameters": {
            "type": "object",
            "properties": {
                "phone_number": {
                    "type": "string",
                    "description": "The phone number of the selected person",
                },
            },
            "required": ["phone_number"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=phone_call)


"""
ToolMate AI Plugin - share

share text or previous generated result on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config, stopSpinning
    import pydoc

    def share(function_args):
        content = config.currentMessages[-1]["content"]
        config.stopSpinning()
        pydoc.pipepager(content, cmd="termux-share -a send")
        return ""

    functionSignature = {
        "examples": [],
        "name": "share",
        "description": f'''Share text or previous generated result with other apps''',
        "parameters": {
            "type": "object",
            "properties": {},
            "required": [],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=share)


"""
ToolMate AI Plugin - ask llama.cpp server

Ask llama.cpp server Model for conversation only; no function calling

[TOOL_CALL]
"""


if not config.isTermux:

    from toolmate import config
    from toolmate.utils.call_llamacppserver import CallLlamaCppServer

    def ask_llamacppserver(function_args):
        config.stopSpinning()
        query = function_args.get("query") # required
        config.currentMessages[-1] = {"role": "user", "content": query}
        completion = CallLlamaCppServer.regularCall(config.currentMessages)
        config.toolmate.streamCompletion(completion, openai=True)
        return ""

    functionSignature = {
        "examples": [
            "Ask Llama.cpp Server",
        ],
        "name": "ask_llamacppserver",
        "description": "Ask Llama.cpp Server to chat or provide information",
        "parameters": {
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=ask_llamacppserver)
    config.inputSuggestions.append("Ask Llama.cpp Server: ")


"""
ToolMate AI Plugin - share file

share file on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config
    import subprocess

    def share_file(function_args):
        file_path = function_args.get("file_path").replace('"', '\\"') # required
        config.stopSpinning()
        cli = f'''termux-share -a send "{file_path}"'''
        subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return ""

    functionSignature = {
        "examples": [
            "share file",
        ],
        "name": "share_file",
        "description": f'''Share a file''',
        "parameters": {
            "type": "object",
            "properties": {
                "file_path": {
                    "type": "string",
                    "description": "The file path of the shared file",
                },
            },
            "required": ["file_path"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=share_file)


"""
ToolMate AI Plugin - send email

send email on Android

[TOOL_CALL]
"""

if config.isTermux:

    from toolmate import config, stopSpinning
    import urllib.parse
    import subprocess

    def send_email(function_args):
        recipient = function_args.get("email", "") # required
        subject = function_args.get("subject", "").replace('"', '\\"') # required
        body = function_args.get("body", "").replace('"', '\\"') # required

        subject = urllib.parse.quote(subject)
        body = urllib.parse.quote(body)

        stopSpinning()

        # e.g. am start -a android.intent.action.SENDTO -d "mailto:john.doe@example.com?subject=Hello&body=How%20are%20you?"
        cli = f'''am start -a android.intent.action.SENDTO -d "mailto:{recipient}?subject={subject}&body={body}"'''
        subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        return ""

    functionSignature = {
        "examples": [
            "send an email",
        ],
        "name": "send_email",
        "description": "Send email",
        "parameters": {
            "type": "object",
            "properties": {
                "email": {
                    "type": "string",
                    "description": "The recipient email address",
                },
                "subject": {
                    "type": "string",
                    "description": "Give a title to the email.",
                },
                "body": {
                    "type": "string",
                    "description": "The body or content of the email.",
                },
            },
            "required": ["email", "subject", "body"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=send_email)


from toolmate import config, get_or_create_collection, add_vector, getFilenamesWithoutExtension, execPythonFile
from toolmate import print2, print3
from pathlib import Path
from prompt_toolkit.completion import FuzzyCompleter, NestedCompleter, ThreadedCompleter
import os, shutil, json, copy
from typing import Callable
if not config.isTermux:
    import chromadb
    from chromadb.config import Settings

class Plugins:

    @staticmethod
    def runPlugins():
        # The following config values can be modified with plugins, to extend functionalities
        #config.pluginsWithFunctionCall = []
        config.aliases = {}
        config.tempChatSystemMessage = ""
        config.predefinedChatSystemMessages = {}
        config.predefinedContexts = {
            "custom": "",
        }
        config.predefinedInstructions = {}
        config.inputSuggestions = [
            "@recommend_tool ", # note: have to keep as the first tool
            "@command ",
            "@append_command ",
            "@append_instruction ",
            "@improve_writing ",
            "@convert_relative_datetime ",
            "@list_current_directory_contents ",
            "@extract_python_code ",
            "@run_python_code ",
        ]
        if config.isTermux:
            config.inputSuggestions.append("@termux ")
        config.outputTransformers = []
        config.deviceInfoPlugins = []
        config.datetimeSensitivePlugins = []
        config.toolFunctionSchemas = {}
        config.toolFunctionMethods = {}
        config.builtinTools = {
            "recommend_tool": "Recommand an appropriate tool in response to a given request",
            #"system": "Apply a predefined system message",
            #"context": "Apply a predefined context",
            "convert_relative_datetime": "Convert relative dates and times in a given instruction to absolute dates and times",
            #"copy_to_clipboard": "Copy a given content to the system clipboard",
            #"paste_from_clipboard": "Retrieve the text content from the system clipboard and paste",
            "extract_python_code": "Extract the python code in a given content",
            "run_python_code": "Extract and run the python code in a given content",
            "list_current_directory_contents": "List the contents in the current directory",
            "command": "Execute a system command",
            "append_command": "Execute a system command with the previous text output appended to it",
            "append_instruction": "Append the previous text output to a given instruction",
            "improve_writing": "Improve the writing of a given content",
        }
        if config.isTermux:
            config.builtinTools["termux"] = "Execute a termux api command"

        pluginFolder = os.path.join(config.toolMateAIFolder, "plugins")
        if config.localStorage:
            customPluginFoler = os.path.join(config.localStorage, "plugins")
            Path(customPluginFoler).mkdir(parents=True, exist_ok=True)
            pluginFolders = (pluginFolder, customPluginFoler)
        else:
            pluginFolders = (pluginFolder,)
        # always run 'search google'
        internetSeraches = "search google"
        script = os.path.join(pluginFolder, "{0}.py".format(internetSeraches))
        execPythonFile(script)
        # always include the following plugins
        requiredPlugins = (
            "chat",
            "auto correct python code",
            "execute computing tasks",
            "execute termux command",
        )
        for i in requiredPlugins:
            if i in config.pluginExcludeList:
                config.pluginExcludeList.remove(i)
        # execute enabled plugins
        for folder in pluginFolders:
            for plugin in getFilenamesWithoutExtension(folder, "py"):
                if not plugin in config.pluginExcludeList:
                    script = os.path.join(folder, "{0}.py".format(plugin))
                    run = execPythonFile(script)
                    if not run:
                        config.pluginExcludeList.append(plugin)
        if internetSeraches in config.pluginExcludeList:
            del config.toolFunctionSchemas["search_google"]

        # tool pattern
        toolNames = "|".join(config.toolFunctionMethods.keys())
        config.toolPattern = f"""@({"|".join(config.builtinTools.keys())}|{toolNames})[ \n]"""
        config.allEnabledTools = list(config.builtinTools.keys())[1:] + list(config.toolFunctionMethods.keys()) # exclude the tool `recommend_tool`
        # input suggestions
        if hasattr(config, "currentMessages"):
            Plugins.buildInputSuggestions()

    # integrate function call plugin
    @staticmethod
    def addFunctionCall(signature: str, method: Callable[[dict], str], deviceInfo=False, datetimeSensitive=False):
        if hasattr(config, "currentMessages"):
            name = signature["name"]
            if not name in config.toolFunctionSchemas: # prevent duplicaiton
                config.toolFunctionSchemas[name] = {key: value for key, value in signature.items() if not key in ("intent", "examples")}
                config.toolFunctionMethods[name] = method
                print3(f"Adding tool: {name}")
                if deviceInfo:
                    config.deviceInfoPlugins.append(name)
                if datetimeSensitive:
                    config.datetimeSensitivePlugins.append(name)
                # input suggestions
                if not name in ("python_qa",):
                    callEntry = f"@{name} "
                    if not callEntry in config.inputSuggestions:
                        config.inputSuggestions.append(callEntry)

    # input suggestions
    @staticmethod
    def buildInputSuggestions():
        nestedSuggestions = {i: None for i in config.actionKeys}
        nestedSuggestions["@chat"] = {f"`{i}` ":None for i in config.predefinedChatSystemMessages}
        for i in config.predefinedContexts:
            if not i.startswith("["):
                nestedSuggestions["@chat"][f"`{i}` "] = None
        for i in config.inputSuggestions:
            if isinstance(i, str):
                nestedSuggestions[i] = None
            elif isinstance(i, dict):
                for iKey, iValue in i.items():
                    nestedSuggestions[iKey] = iValue
        nestedSuggestions = dict(sorted(nestedSuggestions.items()))
        # add config items in developer mode
        nestedSuggestions_developer = copy.deepcopy(nestedSuggestions)
        for i in dir(config):
            if not i.startswith("__"):
                nestedSuggestions_developer[f"config.{i}"] = None
        # completer
        config.completer_user = FuzzyCompleter(ThreadedCompleter(NestedCompleter.from_nested_dict(nestedSuggestions)))
        config.completer_developer = FuzzyCompleter(ThreadedCompleter(NestedCompleter.from_nested_dict(nestedSuggestions_developer)))

    # display available tools
    @staticmethod
    def checkAvailableTools(display=True, includeRequirements=False):
        print("")
        config.toolTextOutput = "# Tools\n"
        tools = copy.deepcopy(config.builtinTools)
        for key, value in config.toolFunctionSchemas.items():
            description = value.get("description", "")
            if includeRequirements:
                try:
                    requirements = value["parameters"].get("required", [])
                except:
                    requirements = []
                if requirements:
                    description += f" (Requirements: {str(requirements)[1:-1]})"
            tools[key] = description
        tools = dict(sorted(tools.items()))
        for key, value in tools.items():
            tool = f"@{key}: {value}"
            config.toolTextOutput += f"\n{tool}"
            if display:
                print3(tool)
        if display:
            print("")



from toolmate import print1, print3, isCommandInstalled
import os

if not isCommandInstalled("ffmpeg"):

    checks = {
        "pkg": "pkg install ffmpeg", # Android Termux
        "brew": "brew install ffmpeg", # on MacOS using Homebrew (https://brew.sh/)
        "choco": "choco install ffmpeg", # on Windows using Chocolatey (https://chocolatey.org/)
        "scoop": "scoop install ffmpeg", # on Windows using Scoop (https://scoop.sh/)
        "apt": "sudo apt update && sudo apt install ffmpeg", # Ubuntu or Debian-based distributions
        "dnf": "sudo dnf install ffmpeg", # Fedora or CentOS
        "pacman": "sudo pacman -Sy ffmpeg", # Arch Linux
        "zypper": "sudo zypper install ffmpeg", # openSUSE
        "yum": "sudo yum localinstall --nogpgcheck https://download1.rpmfusion.org/free/el/rpmfusion-free-release-7.noarch.rpm && sudo yum install ffmpeg", # RHEL, CentOS 7
    }

    for command, commandline in checks.items():
        if isCommandInstalled(command):
            os.system(commandline)
            if isCommandInstalled("ffmpeg"):
                break

    if not isCommandInstalled("ffmpeg"):
        print3("Note: 'ffmpeg' is not installed.")
        print1("It is essential for voice typing with openai whisper offline model, downloading YouTube media, video / audio conversion, etc.")


"""
ToolMate AI Plugin - ask Llama2

Ask Llama2 for information

[TOOL_CALL]
"""

from toolmate import config
from toolmate import print2
from toolmate.ollamachat import OllamaChat
#from toolmate.utils.ollama_models import ollama_models

from toolmate.utils.call_ollama import CallOllama

def ask_llama3_1(function_args):
    model = "llama3.1"
    config.stopSpinning()
    query = function_args.get("query") # required
    config.currentMessages[-1] = {"role": "user", "content": query}
    completion = CallOllama.regularCall(config.currentMessages, chat_model=model)
    config.toolmate.streamCompletion(completion, openai=False)
    if not model == config.ollamaToolModel:
        CallOllama.unloadModels(model)
    return ""

functionSignature = {
    "examples": [
        "Ask Llama3.1",
    ],
    "name": "ask_llama3_1",
    "description": "Ask Llama3.1 to chat or provide information",
    "parameters": {
        "type": "object",
        "properties": {
            "query": {
                "type": "string",
                "description": "The request in detail, including any supplementary information",
            },
        },
        "required": ["query"],
    },
}

config.addFunctionCall(signature=functionSignature, method=ask_llama3_1)
config.inputSuggestions.append("Ask Llama3.1: ")